"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
// available decoders
const codabar = require("./codabar");
const code128 = require("./code-128");
const code39 = require("./code-39");
const code93 = require("./code-93");
const code2of5 = require("./code2of5");
const ean13 = require("./ean-13");
const ean8 = require("./ean-8");
const adaptiveThreshold_1 = require("./utilities/adaptiveThreshold");
const combineAllPossible_1 = require("./utilities/combineAllPossible");
const getImageDataFromSource_1 = require("./utilities/getImageDataFromSource");
const getLines_1 = require("./utilities/getLines");
const isTestEnv = process && process.env.NODE_ENV === 'test';
// TODO: add fuse.js for fuzzy search for bad quality barcodes
var BARCODE_DECODERS;
(function (BARCODE_DECODERS) {
    BARCODE_DECODERS["code-128"] = "code-128";
    BARCODE_DECODERS["code-2of5"] = "code-2of5";
    BARCODE_DECODERS["code-39"] = "code-39";
    BARCODE_DECODERS["code-93"] = "code-93";
    BARCODE_DECODERS["ean-13"] = "ean-13";
    BARCODE_DECODERS["ean-8"] = "ean-8";
    BARCODE_DECODERS["codabar"] = "codabar";
})(BARCODE_DECODERS = exports.BARCODE_DECODERS || (exports.BARCODE_DECODERS = {}));
function isImageLike(object) {
    return object.data && object.width && object.height;
}
function javascriptBarcodeReader({ image, barcode, barcodeType, options, }) {
    return __awaiter(this, void 0, void 0, function* () {
        let decoder;
        switch (barcode) {
            case BARCODE_DECODERS.codabar:
                decoder = codabar.decoder;
                break;
            case BARCODE_DECODERS['code-128']:
                decoder = code128.decoder;
                break;
            case BARCODE_DECODERS['code-39']:
                decoder = code39.decoder;
                break;
            case BARCODE_DECODERS['code-93']:
                decoder = code93.decoder;
                break;
            case BARCODE_DECODERS['code-2of5']:
                decoder = code2of5.decoder;
                break;
            case BARCODE_DECODERS['ean-13']:
                decoder = ean13.decoder;
                break;
            case BARCODE_DECODERS['ean-8']:
                decoder = ean8.decoder;
                break;
            default:
                throw new Error(`Invalid barcode specified. Available decoders: ${BARCODE_DECODERS}.`);
        }
        const useSinglePass = isTestEnv || (options && options.singlePass) || false;
        const imageData = isImageLike(image) ? image : yield getImageDataFromSource_1.getImageDataFromSource(image);
        const width = imageData.width;
        const height = imageData.height;
        const channels = imageData.data.length / (width * height);
        let finalResult = '';
        // apply adaptive threshold
        if (options && options.useAdaptiveThreshold) {
            adaptiveThreshold_1.applyAdaptiveThreshold(imageData.data, width, height);
        }
        // check points for barcode location
        const sPoints = [5, 6, 4, 7, 3, 8, 2, 9, 1];
        const slineStep = Math.round(height / sPoints.length);
        const rowsToScan = Math.min(3, height);
        for (let i = 0; i < sPoints.length; i += 1) {
            const sPoint = sPoints[i];
            const start = channels * width * Math.floor(slineStep * sPoint);
            const end = start + rowsToScan * channels * width;
            const lines = getLines_1.getLines(imageData.data.slice(start, end), width, rowsToScan);
            if (lines.length === 0) {
                if (useSinglePass)
                    throw new Error('Failed to extract barcode!');
                continue;
            }
            // Run the decoder
            const result = decoder(lines, barcodeType);
            if (useSinglePass)
                return result;
            if (!result)
                continue;
            if (!result.includes('?'))
                return result;
            finalResult = combineAllPossible_1.combineAllPossible(finalResult, result);
            if (!finalResult.includes('?'))
                return finalResult;
            if (i === sPoints.length - 1)
                return finalResult;
        }
        throw new Error('Failed to extract barcode!');
    });
}
exports.javascriptBarcodeReader = javascriptBarcodeReader;
//# sourceMappingURL=index.js.map